/*
 *  MOB: A Memory Organization Benchmarking Tool 
 *
 *  Copyright (C) 2000-2001 Josep M. Blanquer and Robert C. Chalmers 
 *
 *  This package is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Josep M. Blanquer <blanquer@cs.ucsb.edu> 
 *          Robert C. Chalmers <robertc@cs.ucsb.edu> 
 *
 */
/**
 * Command line parsing functions.
 **/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "inc/mob.h"
#include "inc/cmdLine.h"
#include "inc/benchmark.h"


/* global variables externally defined */
extern struct globalSystem sys;


/* local function prototypes */
static void parseBenchmarks( char *arg );


/*
  Parse all command line arguments.

  argc - number of arguments
  argv - list of argument strings
 */
void parseCmdLine( int argc, char *argv[] ) {

  int c;

  /* set default values */
  sys.args.program = argv[0];

  /* parse each option and its arguments */
  while( (c = getopt( argc, argv, "b:c:o:r:t:v:ih?" )) != EOF )
    switch (c) {
    case 'b' : 
      parseBenchmarks( optarg );
      break;
    case 'c' :
      sys.args.config = optarg;
      break;
    case 'o' :
      sys.args.output = optarg;
      break;
    case 't' :
      sys.args.trials = atoi( optarg );
      if( sys.args.trials <= 0 )
	printUsage( "trials must be positive", 2 );
      break;
    case 'r' :
      sys.args.runTime = atoi( optarg )*1000;/* convert to usec */
      if( sys.args.runTime <= 0 )
	printUsage( "time must be positive", 2 );
      break;
    case 'v' :
      sys.args.verbosity = atoi( optarg );
      if( sys.args.verbosity < 0 )
	printUsage( "verbosity level must be zero or greater", 2 );
      break;
    case 'i' :
      sys.args.interactive++;
      fprintf( stderr, "Warning: interactive mode is not available.\n" );
      break;
    default :
      printUsage( NULL, 0 );
    }
  
  /* check that no arguments were left on the command line */
  if( optind < argc )
    printUsage( NULL, 2 );
}


/*
  Parse a possible list of benchmark numbers and/or mneumonics.

  arg - benchmark option argument to parse
*/
static void parseBenchmarks( char *arg ) {

  char *mark, *sep = ARG_SEPARATOR;

  /* get pointer to first value */
  mark = strtok( arg, sep );
  /* continue to parse while available */
  while( mark ) {
    /* check whether mark is a number id or mneumonic */
    if( isalpha( (int)mark[0] ) ) {
      /* find mneumonic for individual or bundled benchmarks */
      if( runBenchmarks( mark ) )
      	printUsage( "benchmark not found", 2 );
    } 
    /* ensure number is within range and set to run state */
    else if( runBenchmark( atoi( mark ) ) )
      printUsage( "benchmark out of range", 2 );
    
    /* get next benchmark */
    mark = strtok( NULL, sep );
  }
}


/*
  Print the usage syntax with an optional error message.

  msg - informational message to print in addition to basic usage syntax
  ecode - exit code to return
*/
void printUsage( char *msg, int ecode ) {

  char *base;
  int m;

  /* find basename of current program */
  if( (base = strrchr( sys.args.program, '/' )) )
    base++;
  else
    base = sys.args.program;

  /* print an informational message if passed */
  if( msg )
    fprintf( stderr, "Error: %s\n", msg );
  
  /* print command-line usage and exit with error code */
  fprintf( stderr, "Usage: %s [-b <bm1>[,<bm2>,...]] [-c <config>] [-o output] [-t <trials>] [-r <runtime>] [-v [<verbosity>]] [-i] [-h|?]\n", base );

  /* list each option */
  fprintf( stderr, "Options:\n" );
  fprintf( stderr, "  -b : list of benchmarks numbers or mneumonics\n" );
  fprintf( stderr, "  -c : input configuration file\n" );
  fprintf( stderr, "  -o : output configuration file\n" );
  fprintf( stderr, "  -t : number of times to repeat each test\n" );
  fprintf( stderr, "  -r : minimum time to run one trial\n" );
  fprintf( stderr, "  -v : verbosity level\n" );
  fprintf( stderr, "  -i : run in interactive mode\n" );
  fprintf( stderr, "  -h : print this information\n" );

  /* list available benchmark groups */
  fprintf( stderr, "Bundled Benchmarks:\n" );
  fprintf( stderr, "  all    : run all benchmarks\n" );
  fprintf( stderr, "  caches : run all cache related benchmarks\n" );
  fprintf( stderr, "  tlbs   : run all tlb related benchmarks\n" );

  /* list each available benchmark */
  fprintf( stderr, "Individual Benchmarks:\n" );
  for( m = 0; m < MAX_BENCHMARKS; m++ )
    fprintf( stderr, "  %d. %s\t: %s\n", m, sys.benchmarks[m].mneumonic, sys.benchmarks[m].name );

  /* list each verbosity level */
  fprintf( stderr, "Verbosity Levels\n" );
  fprintf( stderr, "  0 : quiet - nothing is printed\n" );
  fprintf( stderr, "  1 : plot - special output for graphing\n" );
  fprintf( stderr, "  2 : normal - normal printing of results\n" );
  fprintf( stderr, "  3 : warn - include configuration warnings\n" );
  fprintf( stderr, "  4 : inspect - display intermediate test values\n" );
  fprintf( stderr, "  5 : debug - print low-level test parameters\n" );
  fprintf( stderr, "  6 : all - print everything\n" );

  fprintf( stderr, "\n" );

  exit( ecode );
}
